#include "car_physics.hpp"
#include "utils/godot_macros.h"
#include <godot_cpp/classes/curve.hpp>

namespace godot {
void CarPhysics::_bind_methods() {
#define CLASSNAME CarPhysics
    GDPROPERTY_HINTED(oversteer_curve, Variant::OBJECT, PROPERTY_HINT_RESOURCE_TYPE, "Curve");
    GDPROPERTY_HINTED(understeer_curve, Variant::OBJECT, PROPERTY_HINT_RESOURCE_TYPE, "Curve");
}

void CarPhysics::_enter_tree() {
    this->set_use_custom_integrator(true);
}

void CarPhysics::_integrate_forces(PhysicsDirectBodyState3D *state) {
    Basis const basis = this->get_global_basis();

    this->local_velocity.z = Math::move_toward(this->local_velocity.z, this->target_speed, float(ACCELERATION * state->get_step()));
    if(this->local_velocity.z != 0.f) {
        state->set_angular_velocity(basis.get_column(1) * -this->current_steering * Math::sign(this->local_velocity.z));
        this->local_velocity.x = -this->current_steering * 0.1f;
    } else {
        state->set_angular_velocity(Vector3(0,0,0));
        this->local_velocity.x = 0;
    }
    Vector3 world_velocity{
        this->local_velocity.x * basis.get_column(0) +
        this->local_velocity.y * basis.get_column(1) +
        this->local_velocity.z * basis.get_column(2)
    };
    world_velocity.y = 0;
    if(state->get_linear_velocity().y > this->gravity_velocity && state->get_linear_velocity().y <= 0.f)
        this->gravity_velocity = 0.f;
    else
        this->gravity_velocity -= 15.f * state->get_step();
    state->set_linear_velocity(world_velocity + Vector3{0, this->gravity_velocity, 0});
}

void CarPhysics::set_target_speed(float target) {
    this->target_speed = target;
}

float CarPhysics::get_target_speed() const {
    return this->target_speed;
}

void CarPhysics::set_current_steering(float steering) {
    this->current_steering = steering;
}

float CarPhysics::get_current_steering() const {
    return this->current_steering;
}

void CarPhysics::set_brake(bool brake) {
    this->brake = brake;
}

bool CarPhysics::get_brake() const {
    return this->brake;
}

void CarPhysics::set_oversteer_curve(Ref<Curve> curve) {
    this->oversteer_curve = curve;
}

Ref<Curve> CarPhysics::get_oversteer_curve() const {
    return this->oversteer_curve;
}

void CarPhysics::set_understeer_curve(Ref<Curve> curve) {
    this->understeer_curve = curve;
}

Ref<Curve> CarPhysics::get_understeer_curve() const {
    return this->understeer_curve;
}

const float CarPhysics::ACCELERATION{20.f};
const float CarPhysics::BRAKE_FORCE{5.f};
const float CarPhysics::UNDERSTEER_SPEED_MUL{0.f};
const float CarPhysics::OVERSTEER_SPEED_MUL{0.f};
}
